#!/usr/local/gauche/bin/gosh -I.

;
; Funky Fresh Bulletin Board
;
; Copyright (c) 2003, John D. Kilburg
;

(use www.cgi)
(use srfi-27)
(use srfi-19)
(use text.html-lite)
(use text.tree)
(use srfi-1)
(use rfc.cookie)
(use file.util)
(use srfi-13)
(use srfi-14)
(use gauche.regexp)

(use session)
(use ffc)
(use userdb)

;
; Title.  Change this.
;
(define FFBB-NAME "MC^2 Forum")

;
; Alarm!  You must create these directories for stuff to work.  And you
; probably want to change the directories...
;
(define SESSIONDIR "/space/httpd/bulletindb/session")
(define DATADIR "/space/httpd/bulletindb/data")
(define USERPASS "/space/httpd/bulletindb/users")

;
; Something you could easily have to change.
;
(define BULLETIN-SCM-LOCATION "/bulletin/bulletin.scm")

;
; Something you may want to change.
;
(define MAXAGE 3600)    ; session lifetime in seconds

;
; Stuff casual users do not want to change anything below.
;
(define COOKIENAME "BBSESSID")

(define LOGIN-VALUE "Login")
(define NEWTOPIC-VALUE "New Topic")
(define SUBMITTOPIC-VALUE "Submit Topic")
(define LOGOUT-VALUE "Logout")
(define NEWMESSAGE-VALUE "New Message")
(define READTOPIC-VALUE "R")
(define MAINMENU-VALUE "Topic List")
(define TOPICLIST-REFRESH-VALUE "Refresh Topic List")
(define SUBMITMESSAGE-VALUE "Submit Message")
(define NEWUSER-VALUE "New User")

;
; print-error
;
(define (pe . args)
  (for-each
   (lambda (thing) (display thing (current-error-port)))
   args)
  (newline (current-error-port))
  (flush (current-error-port)))

;
; print-flush
;
(define (pf . args)
  (for-each
   (lambda (thing) (display thing))
   args)
  (newline)
  (flush))

;
; generate-topic-id
;
(define (generate-msgid session)
  (format #f "~d-~a"
	  (time->seconds (current-time))
	  (session-get-var session "login")))

;
; header-wif-cookies
;
(define (header-wif-cookies session)
  (tree->string
   (cgi-header
    :cookies (construct-cookie-string
	      (list (list COOKIENAME (session-get-id session)))))))

;
; Footnotes
;
(define (footnotes)
  "<hr />Funky Fresh Bulletin Board.  Copyright (c) 2003.")

;
; get-topicid-parameter
;
; need to sanitize topicid since its part of a path and comes from the
; user.
;
(define (get-topicid-param params)
  (let ((topicid (cgi-get-parameter "topicid" params))
	(tm      #f))
    (set! tm (#/[0-9]\.[0-9]+e[0-9]\-[a-zA-Z_]+/ topicid))
    (if tm
	topicid   ; return the topicid if it looks sane
	#f))) ; return something that will cause trouble...

;
; get-topic-title
;
(define (get-topic-title topicid)
  (with-error-handler
   (lambda (eo) "error-getting-topic")
   (lambda ()
     (with-input-from-port (open-input-file
			    (format #f "~a/~a/info" DATADIR topicid))
       (lambda () (read-line))))))

;
; make-dirlist
;
(define (make-dirlist session)
  (let ((dirlist (list)))
    (for-each
     (lambda (dir)
       (let ((dpath (format #f "~a/~a" DATADIR dir))
	     (dc    (string-split dir "-")))
	 (push! dirlist (list dir
			      (get-topic-title dir)
			      (file-mtime dpath)
			      (cadr dc)))))
     (directory-list DATADIR :children? #t))
    (sort! dirlist (lambda (x y) (> (caddr x) (caddr y))))))

;
; get-post-count
;
; subtract one because of the topic name file.
;
(define (get-post-count topicdir)
  (let ((ap   (format #f "~a/~a" DATADIR topicdir)))
    (- (length (directory-list ap
			       :children? #t :add-path? #t)) 1)))

;
; generate-topiclist
;
(define (generate-topiclist session)
  (html:table
   :width "100%"
   (html:tr
    :bgcolor "#8fff8f"
    (html:td "Read") (html:td "Topic") (html:td "Creator")
    (html:td "Posts") (html:td "Last Post"))
   (let ((flag 0)
	 (msgcolor #f))
     (map
      (lambda (dirinfo)
	(set! flag (- 1 flag))
	(set! msgcolor (if (= flag 0) "#cfcfff" "#cfcfcf"))
	(html:tr
	 :bgcolor msgcolor
	 (html:form
	  :method "POST" :action BULLETIN-SCM-LOCATION
	  (html:input :name "topicid" :type "hidden" :value (car dirinfo))
	  (html:td (html:input :name "submit" :type "submit"
			       :value READTOPIC-VALUE))
	  (html:td (html-escape-string (cadr dirinfo)))
	  (html:td (html-escape-string (cadddr dirinfo)))
	  (html:td (get-post-count (car dirinfo)))
	  (html:td (date->string
		    (time-utc->date
		     (seconds->time
		      (caddr dirinfo))))))))
      (make-dirlist session)))))

;
; generate-topic-html
;
(define (make-msglist session topicid)
  (let ((msglist (list)))
    (for-each
     (lambda (file)
       (if (not (string=? file "info"))
	   (let ((slist (string-split file "-")))
	     (push!
	      msglist
	      (list file (car slist) (cadr slist)
		    (port->string
		     (open-input-file
		      (format #f "~a/~a/~a" DATADIR topicid file))))))))
     (directory-list (format #f "~a/~a" DATADIR topicid) :children? #t))
    (sort! msglist (lambda (x y) (string<? (car x) (car y))))))

(define (generate-topic-html session topicid)
  (html:table
   :width "100%"
   (html:tr
    :bgcolor "#8fff8f"
    (html:td "Poster") (html:td "Message"))
   (let ((flag 0)
	 (msgcolor #f))
     (map
      (lambda (thing)
	(set! flag (- 1 flag))
	(set! msgcolor (if (= flag 0) "#cfcfff" "#cfcfcf"))
	(html:tr
	 :bgcolor msgcolor
	 (html:td :valign "top" (html-escape-string (caddr thing)))
	 (html:td :width "90%"
		  (date->string
		   (time-utc->date
		    (seconds->time
		     (string->number (cadr thing)))))
		  (html:hr)
		  (ffc-expand-string (html-escape-string (cadddr thing))))))
      (make-msglist session topicid)))))

;
; emit-newtopic
;
(define (emit-newtopic session)
  (display (header-wif-cookies session))
  (display
   (tree->string
    (html:html
     (html:head (html:title FFBB-NAME))
     (html:body
      (html:h1 FFBB-NAME)
      (html:h2 "New Topic")
      (html:form
       :method "POST" :action BULLETIN-SCM-LOCATION
       (html:table
	(html:tr
	 (html:td :valign "top" "Subject :")
	 (html:td (html:input :name "topic" :type "input")))
	(html:tr
	 (html:td :valign "top" "Message :")
	 (html:td (html:textarea :name "blather" :cols 60 :rows 12))))
       (html:table
	(html:tr
	 (html:td (html:input :name "submit" :type "submit"
			      :value SUBMITTOPIC-VALUE))
	 (html:td (html:input :name "submit" :type "submit"
			      :value MAINMENU-VALUE))
	 (html:td (html:input :name "submit" :type "submit"
			      :value LOGOUT-VALUE)))))
      (footnotes)))))
  (newline))

;
; emit-newmessage
;
(define (emit-newmessage session topicid)
  (display (header-wif-cookies session))
  (display
   (tree->string
    (html:html
     (html:head (html:title FFBB-NAME))
     (html:body
      (html:h1 FFBB-NAME)
      (html:h2 (format #f "Add Message to ~a" (get-topic-title topicid)))
      (html:form
       :method "POST" :action BULLETIN-SCM-LOCATION
       (html:input :name "topicid" :type "hidden" :value topicid)
       (html:table
	(html:tr
	 (html:td :valign "top" "Message :")
	 (html:td (html:textarea :name "blather" :cols 60 :rows 12))))
       (html:table
	(html:tr
	 (html:td (html:input :name "submit" :type "submit"
			      :value SUBMITMESSAGE-VALUE))
	 (html:td (html:input :name "submit" :type "submit"
			      :value MAINMENU-VALUE))
	 (html:td (html:input :name "submit" :type "submit"
			      :value LOGOUT-VALUE)))))
       (generate-topic-html session topicid)
       (footnotes)))))
  (newline))

;
; emit-main
;
(define (emit-main session)
  (display (header-wif-cookies session))
  (display
   (tree->string
    (html:html
     (html:head (html:title FFBB-NAME))
     (html:body
      (html:h1 FFBB-NAME)
      (html:form
       :method "POST" :action BULLETIN-SCM-LOCATION
       (html:table
	(html:tr
	 (html:td (html:input :name "submit" :type "submit"
			      :value NEWTOPIC-VALUE))
	 (html:td (html:input :name "submit" :type "submit"
			      :value TOPICLIST-REFRESH-VALUE))
	 (html:td (html:input :name "submit" :type "submit"
			      :value LOGOUT-VALUE)))))
      (generate-topiclist session)
      (footnotes)))))
  (newline))

;
; emit-login
;
; Emits an HTML login page.
;
(define (emit-login session message)
  (display (header-wif-cookies session))
  (display
   (tree->string
    (html:html
     (html:head (html:title FFBB-NAME))
     (html:body
      (html:h1 FFBB-NAME)
      (html:h2 "Login")
      message
      (html:p "Enter username and password.  Click Login if you are using an existing username/password.  Click New User if you want to create the username/password.")
      (html:form
       :method "POST" :action BULLETIN-SCM-LOCATION
       (html:table
	(html:tr (html:td "Username:")
		 (html:td (html:input :name "username" :type "text")))
	(html:tr (html:td "Password:")
		 (html:td (html:input :name "password" :type "password"))))
       (html:table
	(html:tr
	 (html:td (html:input :type "submit" :name "submit"
			      :value LOGIN-VALUE))
	 (html:td (html:input :type "submit" :name "submit"
			      :value NEWUSER-VALUE)))))
      (footnotes)))))
  (newline))

;
; emit-logout
;
(define (emit-logout)
  (display (tree->string (cgi-header)))
  (display
   (tree->string
    (html:html
     (html:head (html:title FFBB-NAME))
     (html:body
      (html:h1 FFBB-NAME)
      "You logged out, dawg.  Click "
      (html:a :href BULLETIN-SCM-LOCATION "here") " to log back in."
      (footnotes)))))
  (newline))

;
; emit-topic
;
(define (emit-topic session topicid)
  (display (tree->string (cgi-header)))
  (let ((topic       (get-topic-title topicid)))
    (display
     (tree->string
      (html:html
       (html:head (html:title FFBB-NAME))
       (html:body
	(html:h1 FFBB-NAME)
	(html:h2 (html-escape-string topic))
	(html:form
	 :method "POST" :action BULLETIN-SCM-LOCATION
	 (html:input :type "hidden" :name "topicid" :value topicid)
	 (html:table
	  (html:tr
	   (html:td (html:input :name "submit" :type "submit"
				:value MAINMENU-VALUE))
	   (html:td (html:input :name "submit" :type "submit"
				:value NEWMESSAGE-VALUE))
	   (html:td (html:input :name "submit" :type "submit"
				:value LOGOUT-VALUE)))))
	(generate-topic-html session topicid)
	(footnotes)))))))


;
; Handle page submissions
;
(define (handle-readtopic session params)
  (let ((topicid   (get-topicid-param params)))
    (if (not topicid) (emit-main session) (emit-topic session topicid))))

(define (handle-logout session)
  (session-destroy session)
  (emit-logout))

(define (handle-topic session params)
  (letrec ((topicid (generate-msgid session))
	   (dir     (format #f "~a/~a" DATADIR topicid))
	   (msgpath (format #f "~a/~a" dir topicid))
	   (infpath (format #f "~a/info" dir))
	   (topic   (string-delete
		     (cgi-get-parameter "topic" params) #\newline)))
    (create-directory* dir)
    (with-output-to-port (open-output-file msgpath)
      (lambda ()
	(pf (cgi-get-parameter "blather" params))))
    (with-output-to-port (open-output-file infpath)
      (lambda ()
	(pf topic))))
  (emit-main session))

(define (handle-submitmessage session params)
  (let ((topicid (get-topicid-param params)))
    (if (string? topicid)
	(begin
	  (with-output-to-port (open-output-file
				(format #f "~a/~a/~a" DATADIR topicid
					(generate-msgid session)))
	    (lambda ()
	      (pf (cgi-get-parameter "blather" params))))
;	  (touch-file (format #f "~a/~a" DATADIR topicid))
	  (emit-topic session topicid))
	(emit-main session))))

(define (handle-newmessage session params)
  (let ((topicid (get-topicid-param params)))
    (if (not topicid) (emit-main session) (emit-newmessage session topicid))))

;
; bulletin-board
;
; This is the ``main'' function for the bulletin board.  It dispatches
; submit button events that correspond to the various screens the user
; fiddles with.
;
(define (bulletin-board session params)
  (let ((submit (cgi-get-parameter "submit" params)))
    (if (not (string? submit)) (emit-main session)
	(cond
	 ((string=? submit MAINMENU-VALUE)    (emit-main session))
	 ((string=? submit TOPICLIST-REFRESH-VALUE) (emit-main session))
	 ((string=? submit NEWTOPIC-VALUE)    (emit-newtopic session))
	 ((string=? submit NEWMESSAGE-VALUE)  (handle-newmessage session
								 params))
	 ((string=? submit LOGOUT-VALUE)      (handle-logout session))
	 ((string=? submit SUBMITTOPIC-VALUE) (handle-topic session
							    params))
	 ((string=? submit SUBMITMESSAGE-VALUE) (handle-submitmessage session
								      params))
	 ((string=? submit READTOPIC-VALUE)   (handle-readtopic session
								params))
	 (else (emit-main session)))))
  (session-write session))

;
; username-exists
;
; Returns true if a username exists.
;
(define (username-exists uplist username)
  (call-with-current-continuation
   (lambda (leave-loop)
     (for-each
      (lambda (uinfo)
	(let ((u  (assoc "username" uinfo))
	      (p  (assoc "password" uinfo)))
	  (if (and (list? u) (list? p))
	      (if (string=? username (cadr u))
		  (leave-loop (cadr p))))))
      uplist)
     #f)))

;
; successful-login
;
; Returns true if a username/password is valid.  Returns false if not.
;
(define (successful-login username password)
  (let* ((uplist  (userdb-get-db USERPASS))
	 (pass    (username-exists uplist username)))
    (if (string? pass)
	(string=? password pass)
	#f)))

;
; new-login
;
; Creates a new user if the supplied username and password is not used
; and the username and password meet length and content requirements.
; Returns normally if successful.  Signals an error if failure.
;
(define (new-login username password)
  (let* ((uplist     (userdb-get-db USERPASS))
	 (ulen       (string-length username))
	 (plen       (string-length password)))
    (cond     
     ((number? (string-index
		 username
		 (char-set-union char-set:punctuation
				 char-set:whitespace
				 char-set:symbol)))
      (error "Username contains punctuation/symbols/whitespace which is not allowed."))
     ((number? (string-index
		 password
		 (char-set-union char-set:punctuation
				 char-set:whitespace
				 char-set:symbol)))
      (error "Password contains punctuation/symbols/whitespace which is not allowed."))
     ((string? (username-exists uplist username))
      (error "Username already exists."))
     ((> ulen 24) (error "Username too long."))
     ((> plen 24) (error "Password too long."))
     ((= ulen 0)  (error "Username too short."))
     ((= plen 0)  (error "Password too short."))
     ((number? (string-index username #[\s]))
      (error "Cannot have spaces in username."))
     ((number? (string-index password #[\s]))
      (error "Cannot have spaces in password."))
     ((> (length uplist) 100)
      (error "Too many usernames already."))
     (else
      (begin
	(userdb-save-user (list
			   (list "username" username)
			   (list "password" password)) USERPASS)
	#t)))))

;
; main
;
; Starts the show.  Determine if there is an existing session with a
; valid login.  If there is then run the bulletin board otherwise
; give a login screen.
;
(define (main args)
  (let ((session (session-begin #f SESSIONDIR COOKIENAME))
	(params (cgi-parse-parameters)))
    (session-read session MAXAGE)
    (if (session-get-var session "login")
	(bulletin-board session params)
	(let ((username (cgi-get-parameter "username" params))
	      (password (cgi-get-parameter "password" params))
	      (submit   (cgi-get-parameter "submit"   params)))
	  (if (or (not username) (not password) (not submit))
	      (emit-login session "")
	      (cond
	       ((string=? LOGIN-VALUE submit)
		(if (successful-login username password)
		    ((lambda ()
		       (session-set-var session "login" username)
		       (bulletin-board session params)))
		    (emit-login session "Invalid username/password")))
	       ((string=? NEWUSER-VALUE submit)
		(with-error-handler
		 (lambda (eo)  (emit-login session (slot-ref eo 'message)))
		 (lambda ()
		   (new-login username password)
		   (session-set-var session "login" username)
		   (bulletin-board session params))))
	       (else (emit-login session "Evilness"))))))))
